from __future__ import annotations

import re
from typing import ClassVar, Final, Protocol

from .corr import Corr
from .util import chunked


class DataOverflowError(Exception):
    pass


def to_bytestring(data: str | bytes):
    if isinstance(data, str):
        return data.encode("utf-8")
    return data


def optimal_mode_cls(data: bytes):
    for cls in QRDataNumbers, QRDataAlnums, QRDataBytes:
        if cls.may_represent(data):
            return cls
    raise AssertionError("Can't choose the QRData class")


class QRData(Protocol):
    def write(self, buffer: BitBuffer, version: int): ...

    @classmethod
    def get_mode_size(cls, version: int) -> int: ...

    @classmethod
    def may_represent(cls, data: bytes) -> bool: ...


class QRDataNumbers:
    def __init__(self, data: bytes):
        self.data: Final = data

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def write(self, buffer: BitBuffer, version: int):
        buffer.put(1 << 0, 4)
        buffer.put(len(self.data), self.get_mode_size(version))

        bit_lengths = (-1, 4, 7, 10)
        for chars in chunked(self.data, 3):
            buffer.put(int(chars), bit_lengths[len(chars)])

    @classmethod
    def get_mode_size(cls, version: int):
        if version < 10:
            return 10
        elif version < 27:
            return 12
        else:
            return 14

    @classmethod
    def may_represent(cls, data: bytes) -> bool:
        return data.isdigit()


class QRDataAlnums:

    CHARS: ClassVar = b"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:"
    RE: ClassVar = re.compile(b"^[" + re.escape(CHARS) + rb"]*\Z")

    def __init__(self, data: bytes):
        self.data: Final = data

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def write(self, buffer: BitBuffer, version: int):
        buffer.put(1 << 1, 4)
        buffer.put(len(self.data), self.get_mode_size(version))

        lut = self.CHARS

        for chars in chunked(self.data, 2):
            if len(chars) > 1:
                buffer.put(lut.index(chars[0]) * 45 + lut.index(chars[1]), 11)
            else:
                buffer.put(lut.index(chars), 6)

    @classmethod
    def get_mode_size(cls, version: int):
        if version < 10:
            return 9
        elif version < 27:
            return 11
        else:
            return 13

    @classmethod
    def may_represent(cls, data: bytes) -> bool:
        return cls.RE.match(data) is not None


class QRDataBytes:
    def __init__(self, data: bytes):
        self.data: Final = data

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def write(self, buffer: BitBuffer, version: int):
        buffer.put(1 << 2, 4)
        buffer.put(len(self.data), self.get_mode_size(version))

        for c in self.data:
            buffer.put(c, 8)

    @classmethod
    def get_mode_size(cls, version: int):
        if version < 10:
            return 8
        else:
            return 16

    @classmethod
    def may_represent(cls, data: bytes) -> bool:
        return True


# NOTE: this really should be backed by Python bigint, not bytearray
class BitBuffer:

    __slots__ = ("buffer", "length")

    def __init__(self):
        self.buffer = bytearray()
        self.length = 0

    def __repr__(self):
        return self.buffer.hex()

    def put(self, num: int, length: int):
        for shift in range(length - 1, -1, -1):
            self.put_bit((num >> shift) & 1)

    def put_bit(self, bit: int):
        buf_index = self.length // 8
        if len(self.buffer) <= buf_index:
            self.buffer.append(0)
        if bit:
            self.buffer[buf_index] |= 0x80 >> (self.length % 8)
        self.length += 1


def create_data(
    *data_list: QRData,
    version: int,
    corr: Corr,
):
    buffer = BitBuffer()
    for data in data_list:
        data.write(buffer, version)

    bit_capacity = corr.get_bit_capacity()

    if buffer.length > bit_capacity:
        raise DataOverflowError()

    # Terminate the bits (add up to four 0s).
    for _ in range(min(bit_capacity - buffer.length, 4)):
        buffer.put_bit(0)

    # Delimit the string into 8-bit words, padding with 0s if necessary.
    delimit = buffer.length % 8
    if delimit:
        for _ in range(8 - delimit):
            buffer.put_bit(0)

    # Add special alternating padding bitstrings until buffer is full.
    bytes_to_fill = (bit_capacity - buffer.length) // 8
    for i in range(bytes_to_fill):
        buffer.put(0x11 if i % 2 else 0xEC, 8)

    return corr.apply_to(buffer.buffer)


def iter_optimal_splits(data: bytes, pattern: re.Pattern[bytes]):
    while data:
        match = re.search(pattern, data)
        if not match:
            break
        start, end = match.start(), match.end()
        if start:
            yield False, data[:start]
        yield True, data[start:end]
        data = data[end:]
    if data:
        yield False, data


def make_qrdata(data: str | bytes, cls: type[QRDataAlnums | QRDataNumbers | QRDataBytes] | None = None):
    data = to_bytestring(data)

    if cls is None:
        cls = optimal_mode_cls(data)
    else:
        if not cls.may_represent(data):
            raise ValueError(f"Provided data can not be represented in mode {cls}")
    return cls(data)


def optimal_chunks(data: str | bytes, *, min_chunk=4):
    """
    Returns list of QRData chunks optimized to the data content.

    :param min_chunk: The minimum number of bytes in a row to split as a chunk.
    """

    if min_chunk <= 0:
        raise ValueError("min chunk must be > 0")

    data = to_bytestring(data)
    num_re = rb"\d"
    alnum_re = b"[" + re.escape(QRDataAlnums.CHARS) + b"]"
    if len(data) <= min_chunk:
        num_re = re.compile(b"^" + num_re + b"+$")
        alnum_re = re.compile(b"^" + alnum_re + b"+$")
    else:
        re_repeat = b"{" + str(min_chunk).encode("ascii") + b",}"
        num_re = re.compile(num_re + re_repeat)
        alnum_re = re.compile(alnum_re + re_repeat)

    out: list[QRData] = []
    for is_num, chunk in iter_optimal_splits(data, num_re):
        if is_num:
            out.append(QRDataNumbers(chunk))
        else:
            for is_alnum, sub_chunk in iter_optimal_splits(chunk, alnum_re):
                out.append(QRDataAlnums(sub_chunk) if is_alnum else QRDataBytes(sub_chunk))
    return out



from __future__ import annotations

import zlib
from typing import TYPE_CHECKING

from .util import chunked

if TYPE_CHECKING:
    from .qr import Matrix


def transform(matrix: Matrix, *, border: int, scale=1, colors: tuple[int, int] | None = None):
    if not border and scale == 1 and not colors:
        return matrix.rows

    if not colors:
        colors = (0, 1)

    bg = [colors[0]]

    width = matrix.order * scale + border * 2
    v_border = bg * border
    h_borders = [bg * width] * border

    out: list[list[int]] = []

    out.extend(h_borders)
    for row in matrix.rows:
        transformed: list[int] = []
        transformed.extend(v_border)
        for mod in row:
            color = colors[mod]
            transformed.extend([color] * scale)
        transformed.extend(v_border)
        out.extend([transformed] * scale)
    out.extend(h_borders)

    return out


def as_ascii(matrix: Matrix, *, invert=False, border=0, scale=1):

    codes: dict[tuple[int, ...], str] = {
        (0, 0): "\u00A0",
        (0, 1): "\u2584",
        (1, 0): "\u2580",
        (1, 1): "\u2588",
        (0,): "\u00A0",
        (1,): "\u2580",
    }

    rows = transform(matrix, border=border, colors=(1, 0) if invert else (0, 1), scale=scale)

    out: list[str] = []
    for row_pair in chunked(rows, 2):
        out.append("".join(codes[mod_pair] for mod_pair in zip(*row_pair, strict=False)))
    return out


def _png_block(name: bytes, data: bytes):
    content = name + data
    return [len(data).to_bytes(4), content, zlib.crc32(content).to_bytes(4)]


def as_png(matrix: Matrix, *, invert=False, border: int | None = None, scale=4):

    pixels = bytearray()

    rows = transform(matrix, border=border or 0, scale=scale, colors=(0, 255) if invert else (255, 0))

    for row in rows:
        pixels.append(0)
        pixels.extend(row)

    pixels = zlib.compress(pixels)

    size = len(rows)
    size_word = size.to_bytes(4)

    return b"".join(
        [
            b"\x89PNG\r\n\x1A\n",
            *_png_block(b"IHDR", size_word + size_word + b"\x08\0\0\0\0"),
            *_png_block(b"IDAT", pixels),
            *_png_block(b"IEND", b""),
        ]
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Protocol, cast

if TYPE_CHECKING:
    from .qr import Matrix

# keep it dead simple = let the masks be the function with extra 'code' attribute


class Mask(Protocol):
    code: int

    @classmethod
    def __call__(cls, /, i: int, j: int) -> bool: ...


def _make_mask(code: int, f: Callable[[int, int], bool]) -> Mask:
    f.code = code
    return cast(Mask, f)


MASK0 = _make_mask(0, lambda i, j: (i + j) % 2 == 0)
MASK1 = _make_mask(1, lambda i, j: i % 2 == 0)
MASK2 = _make_mask(2, lambda i, j: j % 3 == 0)
MASK3 = _make_mask(3, lambda i, j: (i + j) % 3 == 0)
MASK4 = _make_mask(4, lambda i, j: ((i // 2) + (j // 3)) % 2 == 0)
MASK5 = _make_mask(5, lambda i, j: (i * j) % 2 + (i * j) % 3 == 0)
MASK6 = _make_mask(6, lambda i, j: ((i * j) % 2 + (i * j) % 3) % 2 == 0)
MASK7 = _make_mask(7, lambda i, j: ((i * j) % 3 + (i + j) % 2) % 2 == 0)


# ---


def choose_mask_pattern(get_matrix: Callable[[Mask], Matrix]):
    candidates: list[tuple[int, int, Mask]] = []

    for mask in MASK0, MASK1, MASK2, MASK3, MASK4, MASK5, MASK6, MASK7:
        matrix = get_matrix(mask)

        penalty = (
            lost_point_level1(matrix)
            + lost_point_level2(matrix)
            + lost_point_level3(matrix)
            + lost_point_level4(matrix)
        )

        # mask.code is here to break ties if penalty is same
        candidates.append((penalty, mask.code, mask))

    return min(candidates)[2]


def lost_point_level1(matrix: Matrix):
    lost_point = 0

    rows = matrix._rows
    rows_count = matrix.order

    rows_range = range(rows_count)
    container = [0] * (rows_count + 1)

    for row in rows_range:
        this_row = rows[row]
        previous_color = this_row[0]
        length = 0
        for col in rows_range:
            if this_row[col] == previous_color:
                length += 1
            else:
                if length >= 5:
                    container[length] += 1
                length = 1
                previous_color = this_row[col]
        if length >= 5:
            container[length] += 1

    for col in rows_range:
        previous_color = rows[0][col]
        length = 0
        for row in rows_range:
            if rows[row][col] == previous_color:
                length += 1
            else:
                if length >= 5:
                    container[length] += 1
                length = 1
                previous_color = rows[row][col]
        if length >= 5:
            container[length] += 1

    lost_point += sum(container[each_length] * (each_length - 2) for each_length in range(5, rows_count + 1))

    return lost_point


def lost_point_level2(matrix: Matrix):

    rows = matrix._rows
    rows_count = matrix.order
    rows_range = range(rows_count - 1)

    lost_point = 0

    for row in rows_range:
        this_row = rows[row]
        next_row = rows[row + 1]
        # use iter() and next() to skip next four-block. e.g.
        # d a f   if top-right a != b bottom-right,
        # c b e   then both abcd and abef won't lost any point.
        col_range_iter = iter(rows_range)
        for col in col_range_iter:
            top_right = this_row[col + 1]
            if top_right != next_row[col + 1]:
                # reduce 33.3% of runtime via next().
                # None: raise nothing if there is no next item.
                next(col_range_iter, None)
            elif top_right != this_row[col]:
                continue
            elif top_right != next_row[col]:
                continue
            else:
                lost_point += 3

    return lost_point


def lost_point_level3(matrix: Matrix):
    # 1 : 1 : 3 : 1 : 1 ratio (dark:light:dark:light:dark) pattern in
    # row/column, preceded or followed by light area 4 modules wide. From ISOIEC.
    # pattern1:     10111010000
    # pattern2: 00001011101

    rows = matrix._rows
    rows_count = matrix.order

    rows_range = range(rows_count)
    rows_range_short = range(rows_count - 10)

    lost_point = 0

    for row in rows_range:
        this_row = rows[row]
        col_range_short_iter = iter(rows_range_short)
        col = 0
        for col in col_range_short_iter:
            if (
                not this_row[col + 1]
                and this_row[col + 4]
                and not this_row[col + 5]
                and this_row[col + 6]
                and not this_row[col + 9]
                and (
                    this_row[col + 0]
                    and this_row[col + 2]
                    and this_row[col + 3]
                    and not this_row[col + 7]
                    and not this_row[col + 8]
                    and not this_row[col + 10]
                    or not this_row[col + 0]
                    and not this_row[col + 2]
                    and not this_row[col + 3]
                    and this_row[col + 7]
                    and this_row[col + 8]
                    and this_row[col + 10]
                )
            ):
                lost_point += 40
            # horspool algorithm.
            # if this_row[col + 10]:
            #   pattern1 shift 4, pattern2 shift 2. So min=2.
            # else:
            #   pattern1 shift 1, pattern2 shift 1. So min=1.
            if this_row[col + 10]:
                next(col_range_short_iter, None)

    for col in rows_range:
        col_range_short_iter = iter(rows_range_short)
        row = 0
        for row in col_range_short_iter:
            if (
                not rows[row + 1][col]
                and rows[row + 4][col]
                and not rows[row + 5][col]
                and rows[row + 6][col]
                and not rows[row + 9][col]
                and (
                    rows[row + 0][col]
                    and rows[row + 2][col]
                    and rows[row + 3][col]
                    and not rows[row + 7][col]
                    and not rows[row + 8][col]
                    and not rows[row + 10][col]
                    or not rows[row + 0][col]
                    and not rows[row + 2][col]
                    and not rows[row + 3][col]
                    and rows[row + 7][col]
                    and rows[row + 8][col]
                    and rows[row + 10][col]
                )
            ):
                lost_point += 40
            if rows[row + 10][col]:
                next(col_range_short_iter, None)

    return lost_point


def lost_point_level4(matrix: Matrix):
    rows = matrix._rows
    rows_count = matrix.order

    dark_count = sum(module for row in rows for module in row if module)
    percent = float(dark_count) / (rows_count**2)
    # Every 5% departure from 50%, rating++
    rating = int(abs(percent * 100 - 50) / 5)
    return rating * 10



from __future__ import annotations

from bisect import bisect_left
from typing import Final, cast

from .corr import Corr, CorrM
from .data import BitBuffer, DataOverflowError, QRData, create_data, make_qrdata
from .mask import Mask, choose_mask_pattern
from .util import bch_type_info, bch_type_number, get_adjust_pattern_pos

CorrectionType = type[Corr]

VERSIONS = range(1, 41)


def choose_version(*data_list: QRData, start_version: int, corr: CorrectionType) -> int:
    """
    Find the minimum size required to fit in the data.
    """

    start_version_mode_sizes = [item.get_mode_size(start_version) for item in data_list]

    buffer = BitBuffer()
    for data in data_list:
        data.write(buffer, start_version)

    need_bits = buffer.length

    # NOTE: Python 3.13+ docs says the bisect is not threadsafe if used on a same sequence.
    version = bisect_left(
        range(VERSIONS.stop), need_bits, lo=start_version, key=lambda ver: corr.get_bit_capacity_for_version(ver)
    )
    if version not in VERSIONS:
        raise DataOverflowError()

    # Now check whether we need more bits for the mode sizes, recursing if
    # our guess was too low
    # XXX: dislike this comparison
    version_mode_sizes = [item.get_mode_size(version) for item in data_list]

    if start_version_mode_sizes != version_mode_sizes:
        version = choose_version(*data_list, start_version=version, corr=corr)
    return version


class Matrix:
    __slots__ = ("version", "order", "_rows")

    def __init__(self, version: int, data: list[list[int | None]] | None = None):
        order = version * 4 + 17
        self.version: Final = version
        self.order: Final = order
        self._rows: Final[list[list[int | None]]]
        if data:
            self._rows = data
        else:
            self._rows = [[None] * order for _ in range(order)]

    def copy(self):
        return Matrix(self.version, [list(row) for row in self._rows])

    def _put_probe_pattern(self, row: int, col: int):
        for r in range(-1, 8):
            if row + r <= -1 or self.order <= row + r:
                continue

            for c in range(-1, 8):
                if col + c <= -1 or self.order <= col + c:
                    continue

                self._rows[row + r][col + c] = int(
                    (0 <= r <= 6 and c in (0, 6)) or (0 <= c <= 6 and r in (0, 6)) or (2 <= r <= 4 and 2 <= c <= 4)
                )

    def _put_all_probe_patterns(self):
        self._put_probe_pattern(0, 0)
        self._put_probe_pattern(self.order - 7, 0)
        self._put_probe_pattern(0, self.order - 7)

    def _put_adjust_pattern(self):
        pos = get_adjust_pattern_pos(self.version)

        for row in pos:
            for col in pos:
                if self._rows[row][col] is not None:
                    continue
                for r in range(-2, 3):
                    for c in range(-2, 3):
                        self._rows[row + r][col + c] = int(
                            r == -2 or r == 2 or c == -2 or c == 2 or (r == 0 and c == 0)
                        )

    def _put_timing_pattern(self):
        for r in range(8, self.order - 8):
            if self._rows[r][6] is not None:
                continue
            self._rows[r][6] = ~r & 1

        for c in range(8, self.order - 8):
            if self._rows[6][c] is not None:
                continue
            self._rows[6][c] = ~c & 1

    def _put_type_info(self, *, test: bool, mask: Mask, corr: Corr):
        bits = bch_type_info((corr.code << 3) | mask.code)

        # vertical
        for i in range(15):
            mod = 0 if test else ((bits >> i) & 1)

            if i < 6:
                self._rows[i][8] = mod
            elif i < 8:
                self._rows[i + 1][8] = mod
            else:
                self._rows[self.order - 15 + i][8] = mod

        # horizontal
        for i in range(15):
            mod = 0 if test else ((bits >> i) & 1)

            if i < 8:
                self._rows[8][self.order - i - 1] = mod
            elif i < 9:
                self._rows[8][15 - i - 1 + 1] = mod
            else:
                self._rows[8][15 - i - 1] = mod

        # fixed module
        self._rows[self.order - 8][8] = 0 if test else 1

    def _put_type_number(self, *, test: bool):
        bits = bch_type_number(self.version)
        order = self.order

        for i in range(18):
            mod = 0 if test else ((bits >> i) & 1)
            self._rows[i // 3][i % 3 + order - 8 - 3] = mod

        for i in range(18):
            mod = 0 if test else ((bits >> i) & 1)
            self._rows[i % 3 + order - 8 - 3][i // 3] = mod

    def _put_data(self, data: bytes, *, mask: Mask):
        inc = -1
        row = self.order - 1
        bit_index = 7
        byte_index = 0

        data_len = len(data)

        for col in range(self.order - 1, 0, -2):
            if col <= 6:
                col -= 1

            col_range = (col, col - 1)

            while True:
                for c in col_range:
                    if self._rows[row][c] is None:
                        bit = 0

                        if byte_index < data_len:
                            bit = (data[byte_index] >> bit_index) & 1

                        bit = bit ^ mask(row, c)

                        self._rows[row][c] = bit
                        bit_index -= 1

                        if bit_index == -1:
                            byte_index += 1
                            bit_index = 7

                row += inc

                if row < 0 or self.order <= row:
                    row -= inc
                    inc = -inc
                    break

    @property
    def rows(self):
        return cast(list[list[int]], self._rows)

    @classmethod
    def blank(cls, version: int):
        matrix = cls(version)
        matrix._put_all_probe_patterns()
        matrix._put_adjust_pattern()
        matrix._put_timing_pattern()
        return matrix


def compile(data_bytes: bytes, *, test: bool, mask: Mask, cache: BlanksCache, version: int, corr: Corr):

    if version in cache:
        matrix = cache[version].copy()
    else:
        matrix = Matrix.blank(version)
        cache[version] = matrix.copy()

    matrix._put_type_info(test=test, mask=mask, corr=corr)

    if version >= 7:
        matrix._put_type_number(test=test)

    matrix._put_data(data_bytes, mask=mask)
    return matrix


def qrcode(
    *data: str | bytes | QRData,
    version: int | None = None,
    error_correction: CorrectionType | None = None,
    mask_pattern: Mask | None = None,
    blanks_cache: BlanksCache | None = None,
):
    data_list = [make_qrdata(item) if isinstance(item, (str, bytes)) else item for item in data]

    if error_correction is None:
        error_correction = CorrM

    if version is None:
        version = choose_version(*data_list, start_version=1, corr=error_correction)
    else:
        assert version in VERSIONS

    corr = error_correction(version)

    data_bytes = create_data(
        *data_list,
        version=version,
        corr=corr,
    )

    if blanks_cache is None:
        blanks_cache = {}

    if mask_pattern is None:

        def _make_test_matrix(mask: Mask):
            return compile(data_bytes, test=True, mask=mask, cache=blanks_cache, version=version, corr=corr)

        mask_pattern = choose_mask_pattern(_make_test_matrix)

    return compile(data_bytes, test=False, mask=mask_pattern, corr=corr, version=version, cache=blanks_cache)


BlanksCache = dict[int, Matrix]



"""Bare headless QR code generator"""

__version__ = "0.1.1"

from .corr import (
    CorrH as CORRECTION_H,
)
from .corr import (
    CorrL as CORRECTION_L,
)
from .corr import (
    CorrM as CORRECTION_M,
)
from .corr import (
    CorrQ as CORRECTION_Q,
)
from .data import (
    optimal_chunks,
)
from .extra import (
    as_ascii,
    as_png,
)
from .mask import MASK0, MASK1, MASK2, MASK3, MASK4, MASK5, MASK6, MASK7
from .qr import (
    BlanksCache,
    qrcode,
)

__all__ = [
    "CORRECTION_L",
    "CORRECTION_M",
    "CORRECTION_Q",
    "CORRECTION_H",
    "MASK0",
    "MASK1",
    "MASK2",
    "MASK3",
    "MASK4",
    "MASK5",
    "MASK6",
    "MASK7",
    "as_ascii",
    "as_png",
    "qrcode",
    "optimal_chunks",
    "BlanksCache",
]
